@@ -8,19 +8,18 @@ module Game.LambdaHack.Client.UI.MonadClientUI
                  , getCachePath
                  )
     -- * Assorted primitives
-  , clientPrintUI, mapStartY, getSession, putSession, displayFrames
+  , clientPrintUI, debugPossiblyPrintUI, getSession, putSession, displayFrames
   , connFrontendFrontKey, setFrontAutoYes, frontendShutdown, printScreen
   , chanFrontend, anyKeyPressed, discardPressedKey, resetPressedKeys
-  , addPressedControlEsc, revCmdMap
-  , getReportUI, getLeaderUI, getArenaUI, viewedLevelUI
-  , leaderTgtToPos, xhairToPos, clearAimMode, scoreToSlideshow, defaultHistory
+  , revCmdMap, getReportUI, getMiniHintAiming, computeChosenLore
+  , getArenaUI, viewedLevelUI, mxhairToPos, xhairToPos, setXHairFromGUI
+  , clearAimMode, getFontSetup, scoreToSlideshow, defaultHistory
   , tellAllClipPS, tellGameClipPS, elapsedSessionTimeGT
-  , resetSessionStart, resetGameStart
-  , partActorLeader, partActorLeaderFun, partPronounLeader
-  , tryRestore, leaderSkillsClientUI
+  , resetSessionStart, resetGameStart, partActorLeader, partPronounLeader
+  , tryRestore, rndToActionUI, tryOpenBrowser
 #ifdef EXPOSE_INTERNAL
     -- * Internal operations
-  , connFrontend, displayFrame, addPressedKey
+  , connFrontend, displayFrame
 #endif
   ) where
 
@@ -28,8 +27,10 @@ import Prelude ()
 
 import Game.LambdaHack.Core.Prelude
 
+import qualified Control.Monad.Trans.State.Strict as St
 import qualified Data.EnumMap.Strict as EM
 import qualified Data.Map.Strict as M
+import qualified Data.Set as S
 import qualified Data.Text as T
 import qualified Data.Text.IO as T
 import           Data.Time.Clock
@@ -37,11 +38,10 @@ import           Data.Time.Clock.POSIX
 import           Data.Time.LocalTime
 import qualified Data.Vector.Unboxed as U
 import qualified NLP.Miniutter.English as MU
-import           System.FilePath
 import           System.IO (hFlush, stdout)
+import           Web.Browser (openBrowser)
 
 import           Game.LambdaHack.Client.Bfs
-import           Game.LambdaHack.Client.ClientOptions
 import           Game.LambdaHack.Client.CommonM
 import           Game.LambdaHack.Client.MonadClient
 import           Game.LambdaHack.Client.State
@@ -49,20 +49,22 @@ import           Game.LambdaHack.Client.UI.ActorUI
 import           Game.LambdaHack.Client.UI.Content.Input
 import           Game.LambdaHack.Client.UI.Content.Screen
 import           Game.LambdaHack.Client.UI.ContentClientUI
+import           Game.LambdaHack.Client.UI.EffectDescription
 import           Game.LambdaHack.Client.UI.Frame
-import           Game.LambdaHack.Client.UI.Frontend
 import qualified Game.LambdaHack.Client.UI.Frontend as Frontend
 import qualified Game.LambdaHack.Client.UI.HumanCmd as HumanCmd
 import qualified Game.LambdaHack.Client.UI.Key as K
 import           Game.LambdaHack.Client.UI.Msg
+import           Game.LambdaHack.Client.UI.Overlay
 import           Game.LambdaHack.Client.UI.SessionUI
 import           Game.LambdaHack.Client.UI.Slideshow
 import           Game.LambdaHack.Client.UI.UIOptions
 import           Game.LambdaHack.Common.Actor
 import           Game.LambdaHack.Common.ActorState
+import           Game.LambdaHack.Common.ClientOptions
 import           Game.LambdaHack.Common.Faction
-import           Game.LambdaHack.Common.File
 import qualified Game.LambdaHack.Common.HighScore as HighScore
+import           Game.LambdaHack.Common.Item
 import           Game.LambdaHack.Common.Kind
 import           Game.LambdaHack.Common.Misc
 import           Game.LambdaHack.Common.MonadStateRead
@@ -72,24 +74,29 @@ import qualified Game.LambdaHack.Common.Save as Save
 import           Game.LambdaHack.Common.State
 import           Game.LambdaHack.Common.Time
 import           Game.LambdaHack.Common.Types
+import           Game.LambdaHack.Content.FactionKind
 import           Game.LambdaHack.Content.ModeKind
-import           Game.LambdaHack.Content.RuleKind
-import qualified Game.LambdaHack.Definition.Ability as Ability
-import           Game.LambdaHack.Definition.Defs
+import           Game.LambdaHack.Core.Random
 
 -- Assumes no interleaving with other clients, because each UI client
 -- in a different terminal/window/machine.
 clientPrintUI :: MonadClientUI m => Text -> m ()
 clientPrintUI t = liftIO $ do
-  T.hPutStrLn stdout t
+  T.hPutStr stdout $! t <> "\n"  -- hPutStrLn not atomic enough
   hFlush stdout
 
--- | The row where the dungeon map starts.
-mapStartY :: Y
-mapStartY = 1
+debugPossiblyPrintUI :: MonadClientUI m => Text -> m ()
+debugPossiblyPrintUI t = do
+  sdbgMsgCli <- getsClient $ sdbgMsgCli . soptions
+  when sdbgMsgCli $ liftIO $ do
+    T.hPutStr stdout $! t <> "\n"  -- hPutStrLn not atomic enough
+    hFlush stdout
 
 -- | The monad that gives the client access to UI operations,
--- but not to modifying client state.
+-- but not to modifying client state, except for the client-side pointman
+-- (as opposed to pointman stores in faction data in main game state),
+-- which is more of a UI concept, but is shared with AI to be able
+-- to keep it when switching AI on/off and to save on typing.
 class MonadClientRead m => MonadClientUI m where
   getsSession :: (SessionUI -> a) -> m a
   modifySession :: (SessionUI -> SessionUI) -> m ()
@@ -104,104 +111,146 @@ putSession :: MonadClientUI m => SessionUI -> m ()
 putSession s = modifySession (const s)
 
 -- | Write a UI request to the frontend and read a corresponding reply.
-connFrontend :: MonadClientUI m => FrontReq a -> m a
+connFrontend :: MonadClientUI m => Frontend.FrontReq a -> m a
 connFrontend req = do
-  ChanFrontend f <- getsSession schanF
+  Frontend.ChanFrontend f <- getsSession schanF
   liftIO $ f req
 
 displayFrame :: MonadClientUI m => Maybe Frame -> m ()
 displayFrame mf = do
   frame <- case mf of
-    Nothing -> return $! FrontDelay 1
+    Nothing -> return $! Frontend.FrontDelay 1
     Just fr -> do
       modifySession $ \cli -> cli {snframes = snframes cli + 1}
-      return $! FrontFrame fr
+      return $! Frontend.FrontFrame fr
   connFrontend frame
 
 -- | Push frames or delays to the frame queue. The frames depict
 -- the @lid@ level.
-displayFrames :: MonadClientUI m => LevelId -> PreFrames -> m ()
+displayFrames :: MonadClientUI m => LevelId -> PreFrames3 -> m ()
+displayFrames _ [] = return ()
 displayFrames lid frs = do
-  let frames = case frs of
+  let framesRaw = case frs of
         [] -> []
-        [Just (bfr, ffr)] -> [Just (FrameBase $ U.unsafeThaw bfr, ffr)]
+        [Just ((bfr, ffr), (ovProp, ovSquare, ovMono))] ->
+          [Just ( (FrameBase $ U.unsafeThaw bfr, ffr)
+                , (ovProp, ovSquare, ovMono) )]
         _ ->
           -- Due to the frames coming from the same base frame,
           -- we have to copy it to avoid picture corruption.
-          map (fmap $ \(bfr, ffr) -> (FrameBase $ U.thaw bfr, ffr)) frs
-  mapM_ displayFrame frames
-  -- Can be different than @blid b@, e.g., when our actor is attacked
-  -- on a remote level.
+          map (fmap $ \((bfr, ffr), (ovProp, ovSquare, ovMono)) ->
+                ((FrameBase $ U.thaw bfr, ffr), (ovProp, ovSquare, ovMono))) frs
+  -- If display level different than the man viewed level,
+  -- e.g., when our actor is attacked on a remote level,
+  -- then pad with tripple delay to give more time to see the remote frames(s).
   lidV <- viewedLevelUI
-  when (lidV == lid) $
-    modifySession $ \sess -> sess {sdisplayNeeded = False}
+  frames <- if lidV == lid
+            then do
+              modifySession $ \sess -> sess { sdisplayNeeded = False
+                                            , sturnDisplayed = True }
+              return framesRaw
+            else return $ framesRaw ++ [Nothing, Nothing, Nothing]
+  mapM_ displayFrame frames
 
--- | Write 'FrontKey' UI request to the frontend, read the reply,
+-- | Write 'Frontend.FrontKey' UI request to the frontend, read the reply,
 -- set pointer, return key.
-connFrontendFrontKey :: MonadClientUI m => [K.KM] -> PreFrame -> m K.KM
-connFrontendFrontKey frontKeyKeys (bfr, ffr) = do
-  let frontKeyFrame = (FrameBase $ U.unsafeThaw bfr, ffr)
-  kmp <- connFrontend $ FrontKey frontKeyKeys frontKeyFrame
-  modifySession $ \sess -> sess {spointer = K.kmpPointer kmp}
-  return $! K.kmpKeyMod kmp
+connFrontendFrontKey :: MonadClientUI m => [K.KM] -> PreFrame3 -> m K.KM
+connFrontendFrontKey frontKeyKeys ((bfr, ffr), (ovProp, ovSquare, ovMono)) = do
+  let frontKeyFrame =
+        ((FrameBase $ U.unsafeThaw bfr, ffr), (ovProp, ovSquare, ovMono))
+  sautoYes <- getsSession sautoYes
+  if sautoYes && (null frontKeyKeys || K.spaceKM `elem` frontKeyKeys) then do
+    connFrontend $ Frontend.FrontFrame frontKeyFrame
+    return K.spaceKM
+  else do
+    kmp <- connFrontend $ Frontend.FrontKey frontKeyKeys frontKeyFrame
+    modifySession $ \sess -> sess {spointer = K.kmpPointer kmp}
+    return $! K.kmpKeyMod kmp
 
 setFrontAutoYes :: MonadClientUI m => Bool -> m ()
-setFrontAutoYes b = connFrontend $ FrontAutoYes b
+setFrontAutoYes b = modifySession $ \sess -> sess {sautoYes = b}
 
 frontendShutdown :: MonadClientUI m => m ()
-frontendShutdown = connFrontend FrontShutdown
+frontendShutdown = connFrontend Frontend.FrontShutdown
 
 printScreen :: MonadClientUI m => m ()
-printScreen = connFrontend FrontPrintScreen
+printScreen = connFrontend Frontend.FrontPrintScreen
 
 -- | Initialize the frontend chosen by the player via client options.
 chanFrontend :: MonadClientUI m
-             => ScreenContent -> ClientOptions -> m ChanFrontend
+             => ScreenContent -> ClientOptions -> m Frontend.ChanFrontend
 chanFrontend coscreen soptions =
   liftIO $ Frontend.chanFrontendIO coscreen soptions
 
 anyKeyPressed :: MonadClientUI m => m Bool
-anyKeyPressed = connFrontend FrontPressed
+anyKeyPressed = connFrontend Frontend.FrontPressed
 
 discardPressedKey :: MonadClientUI m => m ()
-discardPressedKey = connFrontend FrontDiscardKey
+discardPressedKey = connFrontend Frontend.FrontDiscardKey
 
 resetPressedKeys :: MonadClientUI m => m ()
-resetPressedKeys = connFrontend FrontResetKeys
+resetPressedKeys = connFrontend Frontend.FrontResetKeys
 
-addPressedKey :: MonadClientUI m => K.KMP -> m ()
-addPressedKey = connFrontend . FrontAdd
-
-addPressedControlEsc :: MonadClientUI m => m ()
-addPressedControlEsc = addPressedKey K.KMP { K.kmpKeyMod = K.controlEscKM
-                                           , K.kmpPointer = originPoint }
-
-revCmdMap :: MonadClientUI m => m (K.KM -> HumanCmd.HumanCmd -> K.KM)
+revCmdMap :: MonadClientUI m => m (HumanCmd.HumanCmd -> K.KM)
 revCmdMap = do
   CCUI{coinput=InputContent{brevMap}} <- getsSession sccui
-  let revCmd dflt cmd = case M.lookup cmd brevMap of
-        Nothing -> dflt
+  let revCmd cmd = case M.lookup cmd brevMap of
+        Nothing -> K.undefinedKM
         Just (k : _) -> k
         Just [] -> error $ "" `showFailure` brevMap
   return revCmd
 
-getReportUI :: MonadClientUI m => m Report
-getReportUI = do
+getReportUI :: MonadClientUI m => Bool -> m Report
+getReportUI insideMenu = do
+  saimMode <- getsSession saimMode
   sUIOptions <- getsSession sUIOptions
   report <- getsSession $ newReport . shistory
-  side <- getsClient sside
-  fact <- getsState $ (EM.! side) . sfactionD
-  let underAI = isAIFact fact
-      mem = EM.fromList <$> uMessageColors sUIOptions
-      promptAI = toMsg mem MsgPrompt "[press any key for main menu]"
-  return $! if underAI then consReport promptAI report else report
+  sreqDelay <- getsSession sreqDelay
+  miniHintAiming <- getMiniHintAiming
+  -- Different from ordinary tutorial hints in that shown more than once.
+  let detailAtDefault = (detailLevel <$> saimMode) == Just defaultDetailLevel
+      detailMinimal = (detailLevel <$> saimMode) == Just minBound
+      prefixColors = uMessageColors sUIOptions
+      promptAim = toMsgShared prefixColors MsgPromptGeneric
+                              (miniHintAiming <> "\n")
+      promptDelay = toMsgShared prefixColors MsgPromptAction
+                                "<press any key to regain control>"
+  return $! if | not insideMenu && detailAtDefault && not detailMinimal ->
+                   consReport promptAim report
+               | sreqDelay == ReqDelayAlarm && not insideMenu ->
+                   consReport promptDelay report
+               | otherwise -> report
 
-getLeaderUI :: MonadClientUI m => m ActorId
-getLeaderUI = do
-  cli <- getClient
-  case sleader cli of
-    Nothing -> error $ "leader expected but not found" `showFailure` cli
-    Just leader -> return leader
+getMiniHintAiming :: MonadClientUI m => m Text
+getMiniHintAiming = do
+  saimMode <- getsSession saimMode
+  (inhabitants, embeds) <-
+    if isJust saimMode then computeChosenLore else return ([], [])
+  sreqDelay <- getsSession sreqDelay
+  mleader <- getsClient sleader
+  let loreCommandAvailable = not (null inhabitants && null embeds)
+                             && isJust mleader
+  -- Here we assume newbies don't override default keys.
+  return $! T.unwords $ concat
+    [ ["Aiming mode:"]
+    , ["'~' for lore," | loreCommandAvailable ]
+    , ["'f' to fling," | sreqDelay /= ReqDelayHandled]
+    , [if loreCommandAvailable && sreqDelay /= ReqDelayHandled
+       then "SPACE or RMB to hush,"  -- shorter, because less space left
+       else "SPACE or RMB to cycle detail,"]
+    , ["ESC to cancel."] ]
+
+computeChosenLore :: MonadClientUI m
+                  => m ([(ActorId, Actor)], [(ItemId, ItemQuant)])
+computeChosenLore = do
+  side <- getsClient sside
+  xhairPos <- xhairToPos
+  lidV <- viewedLevelUI
+  let isOurs (_, b) = bfid b == side
+  inhabitants0 <- getsState $ filter (not . isOurs)
+                              . posToAidAssocs xhairPos lidV
+  embeds0 <- getsState $ EM.assocs . getEmbedBag lidV xhairPos
+  return (inhabitants0, embeds0)
 
 getArenaUI :: MonadClientUI m => m LevelId
 getArenaUI = do
@@ -228,25 +277,27 @@ viewedLevelUI = do
   saimMode <- getsSession saimMode
   return $! maybe arena aimLevelId saimMode
 
-leaderTgtToPos :: MonadClientUI m => m (Maybe Point)
-leaderTgtToPos = do
-  lidV <- viewedLevelUI
-  mleader <- getsClient sleader
-  case mleader of
-    Nothing -> return Nothing
-    Just aid -> do
-      mtgt <- getsClient $ getTarget aid
-      getsState $ aidTgtToPos aid lidV mtgt
-
-xhairToPos :: MonadClientUI m => m (Maybe Point)
-xhairToPos = do
+mxhairToPos :: MonadClientUI m => m (Maybe Point)
+mxhairToPos = do
   lidV <- viewedLevelUI
   mleader <- getsClient sleader
   sxhair <- getsSession sxhair
-  case mleader of
-    Nothing -> return Nothing  -- e.g., when game start and no leader yet
-    Just aid -> getsState $ aidTgtToPos aid lidV sxhair
-                  -- e.g., xhair on another level
+  getsState $ aidTgtToPos mleader lidV sxhair
+
+xhairToPos :: MonadClientUI m => m Point
+xhairToPos = do
+  mxhairPos <- mxhairToPos
+  mleader <- getsClient sleader
+  fallback <- case mleader of
+    Nothing -> return originPoint
+    Just leader -> getsState $ bpos . getActorBody leader
+  return $! fromMaybe fallback mxhairPos
+
+setXHairFromGUI :: MonadClientUI m => Maybe Target -> m ()
+setXHairFromGUI xhair2 = do
+  xhair0 <- getsSession sxhair
+  modifySession $ \sess -> sess {sxhairGoTo = Nothing}
+  when (xhair0 /= xhair2) $ modifySession $ \sess -> sess {sxhair = xhair2}
 
 -- If aim mode is exited, usually the player had the opportunity to deal
 -- with xhair on a foe spotted on another level, so now move xhair
@@ -254,19 +305,31 @@ xhairToPos = do
 clearAimMode :: MonadClientUI m => m ()
 clearAimMode = do
   lidVOld <- viewedLevelUI  -- not in aiming mode at this point
-  mxhairPos <- xhairToPos  -- computed while still in aiming mode
+  xhairPos <- xhairToPos  -- computed while still in aiming mode
   modifySession $ \sess -> sess {saimMode = Nothing}
   lidV <- viewedLevelUI  -- not in aiming mode at this point
   when (lidVOld /= lidV) $ do
-    leader <- getLeaderUI
-    lpos <- getsState $ bpos . getActorBody leader
     sxhairOld <- getsSession sxhair
-    let xhairPos = fromMaybe lpos mxhairPos
-        sxhair = case sxhairOld of
+    let sxhair = case sxhairOld of
           Just TPoint{} -> Just $ TPoint TUnknown lidV xhairPos
             -- the point is possibly unknown on this level; unimportant anyway
           _ -> sxhairOld
-    modifySession $ \sess -> sess {sxhair}
+    setXHairFromGUI sxhair
+
+-- We can't support setup @FontSetup SquareFont MonoFont MonoFont@
+-- at this time, because the mono layer needs to overwrite the prop layer
+-- and so has to be distinct even if the underlying font is mono for both.
+getFontSetup :: MonadClientUI m => m FontSetup
+getFontSetup = do
+  soptions@ClientOptions{schosenFontset, sfontsets} <- getsClient soptions
+  let chosenFontsetID = fromJust schosenFontset
+      chosenFontset = case lookup chosenFontsetID sfontsets of
+        Nothing -> error $ "Fontset not defined in config file"
+                           `showFailure` chosenFontsetID
+        Just fs -> fs
+      multiFont = Frontend.frontendName soptions == "sdl"
+                  && not (T.null (fontPropRegular chosenFontset))
+  return $! if multiFont then multiFontSetup else singleFontSetup
 
 scoreToSlideshow :: MonadClientUI m => Int -> Status -> m Slideshow
 scoreToSlideshow total status = do
@@ -284,9 +347,6 @@ scoreToSlideshow total status = do
   let fact = factionD EM.! fid
       table = HighScore.getTable gameModeId scoreDict
       gameModeName = mname gameMode
-      chal | fhasUI $ gplayer fact = curChalSer
-           | otherwise = curChalSer
-                           {cdiff = difficultyInverse (cdiff curChalSer)}
       theirVic (fi, fa) | isFoe fid fact fi
                           && not (isHorrorFact fa) = Just $ gvictims fa
                         | otherwise = Nothing
@@ -295,11 +355,13 @@ scoreToSlideshow total status = do
                       | otherwise = Nothing
       ourVictims = EM.unionsWith (+) $ mapMaybe ourVic $ EM.assocs factionD
       (worthMentioning, (ntable, pos)) =
-        HighScore.register table total dungeonTotal time status date chal
+        HighScore.register table total dungeonTotal time status date curChalSer
                            (T.unwords $ tail $ T.words $ gname fact)
                            ourVictims theirVictims
-                           (fhiCondPoly $ gplayer fact)
-      sli = highSlideshow rwidth (rheight - 1) ntable pos gameModeName tz
+                           (fhiCondPoly $ gkind fact)
+  fontSetup <- getFontSetup
+  let sli = highSlideshow fontSetup False rwidth (rheight - 1) ntable pos
+                          gameModeName tz
   return $! if worthMentioning
             then sli
             else emptySlideshow
@@ -307,15 +369,20 @@ scoreToSlideshow total status = do
 defaultHistory :: MonadClientUI m => m History
 defaultHistory = do
   sUIOptions <- getsSession sUIOptions
+  curTutorial <- getsSession scurTutorial
+  overrideTut <- getsSession soverrideTut
+  let displayHints = fromMaybe curTutorial overrideTut
   liftIO $ do
     utcTime <- getCurrentTime
     timezone <- getTimeZone utcTime
     let curDate = T.pack $ take 19 $ show $ utcToLocalTime timezone utcTime
         emptyHist = emptyHistory $ uHistoryMax sUIOptions
-        mem = EM.fromList <$> uMessageColors sUIOptions
-        msg = toMsg mem MsgAdmin
+        msg = toMsgShared (uMessageColors sUIOptions) MsgBookKeeping
               $ "History log started on " <> curDate <> "."
-    return $! fst $ addToReport emptyHist msg 0 timeZero
+        -- Tuturial hints from initial message can be repeated.
+        (_, nhistory, _) =
+          addToReport S.empty displayHints False emptyHist msg timeZero
+    return nhistory
 
 tellAllClipPS :: MonadClientUI m => m ()
 tellAllClipPS = do
@@ -330,8 +397,8 @@ tellAllClipPS = do
     let time = absoluteTimeAdd allTime gtime
         nframes = allNframes + gnframes
         diff = fromRational $ toRational $ curPOSIX - sstartPOSIX
-        cps = fromIntegral (timeFit time timeClip) / diff :: Double
-        fps = fromIntegral nframes / diff :: Double
+        cps = intToDouble (timeFit time timeClip) / diff
+        fps = intToDouble nframes / diff
     clientPrintUI $
       "Session time:" <+> tshow diff <> "s; frames:" <+> tshow nframes <> "."
       <+> "Average clips per second:" <+> tshow cps <> "."
@@ -348,19 +415,24 @@ tellGameClipPS = do
       time <- getsState stime
       nframes <- getsSession snframes
       let diff = fromRational $ toRational $ curPOSIX - sgstartPOSIX
-          cps = fromIntegral (timeFit time timeClip) / diff :: Double
-          fps = fromIntegral nframes / diff :: Double
+          cps = intToDouble (timeFit time timeClip) / diff
+          fps = intToDouble nframes / diff
       -- This means: "Game portion after last reload time:...".
       clientPrintUI $
         "Game time:" <+> tshow diff <> "s; frames:" <+> tshow nframes <> "."
         <+> "Average clips per second:" <+> tshow cps <> "."
         <+> "Average FPS:" <+> tshow fps <> "."
 
-elapsedSessionTimeGT :: MonadClientUI m => Int -> m Bool
-elapsedSessionTimeGT stopAfter = do
+-- TODO: for speed and resolution use
+-- https://hackage.haskell.org/package/chronos
+-- or the number_of_nanonseconds functionality
+-- in Data.Time.Clock.System, once it arrives there
+elapsedSessionTimeGT :: MonadClientRead m => POSIXTime -> Int -> m Bool
+elapsedSessionTimeGT sstartPOSIX stopAfter = do
   current <- liftIO getPOSIXTime
-  sstartPOSIX <- getsSession sstart
-  return $! fromIntegral stopAfter + sstartPOSIX <= current
+  return $! (fromIntegralWrap :: Int -> NominalDiffTime) stopAfter
+            + sstartPOSIX
+            <= current
 
 resetSessionStart :: MonadClientUI m => m ()
 resetSessionStart = do
@@ -373,17 +445,11 @@ resetGameStart = do
   sgstart <- liftIO getPOSIXTime
   time <- getsState stime
   nframes <- getsSession snframes
-  modifySession $ \cli ->
-    cli { sgstart
-        , sallTime = absoluteTimeAdd (sallTime cli) time
+  modifySession $ \sess ->
+    sess { sgstart
+        , sallTime = absoluteTimeAdd (sallTime sess) time
         , snframes = 0
-        , sallNframes = sallNframes cli + nframes }
-
-partActorLeaderCommon :: Maybe ActorId -> ActorUI -> Actor -> ActorId -> MU.Part
-partActorLeaderCommon mleader bUI b aid = case mleader of
-  Just leader | aid == leader -> "you"
-  _ | bhp b <= 0 -> MU.Phrase ["the fallen", partActor bUI]
-  _ -> partActor bUI
+        , sallNframes = sallNframes sess + nframes }
 
 -- | The part of speech describing the actor or the "you" pronoun if he is
 -- the leader of the observer's faction.
@@ -392,15 +458,12 @@ partActorLeader aid = do
   mleader <- getsClient sleader
   bUI <- getsSession $ getActorUI aid
   b <- getsState $ getActorBody aid
-  return $! partActorLeaderCommon mleader bUI b aid
-
-partActorLeaderFun :: MonadClientUI m => m (ActorId -> MU.Part)
-partActorLeaderFun = do
-  mleader <- getsClient sleader
-  sess <- getSession
-  s <- getState
-  return $! \aid ->
-    partActorLeaderCommon mleader (getActorUI aid sess) (getActorBody aid s) aid
+  return $! case mleader of
+    Just leader | aid == leader -> "you"
+    _ | bhp b <= 0
+        && not (bproj b) ->  -- avoid "the fallen falling" projectiles
+      MU.Phrase ["the fallen", partActor bUI]
+    _ -> partActor bUI
 
 -- | The part of speech with the actor's pronoun or "you" if a leader
 -- of the client's faction.
@@ -415,21 +478,22 @@ partPronounLeader aid = do
 -- | Try to read saved client game state from the file system.
 tryRestore :: MonadClientUI m => m (Maybe (StateClient, Maybe SessionUI))
 tryRestore = do
-  cops@COps{corule} <- getsState scops
-  bench <- getsClient $ sbenchmark . soptions
-  if bench then return Nothing
+  COps{corule} <- getsState scops
+  clientOptions <- getsClient soptions
+  if sbenchmark clientOptions then return Nothing
   else do
     side <- getsClient sside
     prefix <- getsClient $ ssavePrefixCli . soptions
-    let fileName = prefix <> Save.saveNameCli cops side
-    res <- liftIO $ Save.restoreGame cops fileName
-    let cfgUIName = rcfgUIName corule
-        content = rcfgUIDefault corule
-    dataDir <- liftIO appDataDir
-    liftIO $ tryWriteFile (dataDir </> cfgUIName) content
-    return res
+    let fileName = prefix <> Save.saveNameCli corule side
+    liftIO $ Save.restoreGame corule clientOptions fileName
 
-leaderSkillsClientUI :: MonadClientUI m => m Ability.Skills
-leaderSkillsClientUI = do
-  leader <- getLeaderUI
-  getsState $ getActorMaxSkills leader
+-- | Invoke pseudo-random computation with the generator kept in the session.
+rndToActionUI :: MonadClientUI m => Rnd a -> m a
+rndToActionUI r = do
+  gen1 <- getsSession srandomUI
+  let (a, gen2) = St.runState r gen1
+  modifySession $ \sess -> sess {srandomUI = gen2}
+  return a
+
+tryOpenBrowser :: MonadClientUI m => String -> m Bool
+tryOpenBrowser address = liftIO $ openBrowser address
