@@ -1,4 +1,3 @@
-{-# LANGUAGE TupleSections #-}
 -- | Operations on the 'Actor' type, and related, that need the 'State' type,
 -- but not our custom monad types.
 module Game.LambdaHack.Common.ActorState
@@ -6,20 +5,17 @@ module Game.LambdaHack.Common.ActorState
   , fidActorRegularIds, foeRegularAssocs, foeRegularList
   , friendRegularAssocs, friendRegularList, bagAssocs, bagAssocsK
   , posToBig, posToBigAssoc, posToProjs, posToProjAssocs
-  , posToAids, posToAidAssocs
-  , calculateTotal, itemPrice, mergeItemQuant, findIid
-  , combinedGround, combinedOrgan, combinedEqp, combinedInv
-  , combinedItems, combinedFromLore
+  , posToAids, posToAidAssocs, calculateTotal, itemPrice, findIid, combinedItems
   , getActorBody, getActorMaxSkills, actorCurrentSkills, canTraverse
-  , getCarriedAssocsAndTrunk, getCarriedIidCStore, getContainerBag
-  , getFloorBag, getEmbedBag, getBodyStoreBag
-  , mapActorItems_, getActorAssocs, getActorAssocsK
-  , memActor, getLocalTime, regenCalmDelta, actorInAmbient, canDeAmbientList
+  , getCarriedAssocsAndTrunk, getContainerBag
+  , getFloorBag, getEmbedBag, getBodyStoreBag, getFactionStashBag
+  , mapActorItems_, getActorAssocs
+  , memActor, getLocalTime, regenCalmDelta, actorInAmbient
   , dispEnemy, itemToFull, fullAssocs, kitAssocs
   , getItemKindId, getIidKindId, getItemKind, getIidKind
   , getItemKindIdServer, getIidKindIdServer, getItemKindServer, getIidKindServer
-  , lidFromC, posFromC, anyFoeAdj, adjacentBigAssocs, adjacentProjAssocs
-  , armorHurtBonus, inMelee
+  , tileAlterable, lidFromC, posFromC, anyFoeAdj, anyHarmfulFoeAdj
+  , adjacentBigAssocs, adjacentProjAssocs, armorHurtBonus, inMelee
   ) where
 
 import Prelude ()
@@ -37,15 +33,15 @@ import           Game.LambdaHack.Common.Item
 import           Game.LambdaHack.Common.Kind
 import           Game.LambdaHack.Common.Level
 import           Game.LambdaHack.Common.Misc
+import           Game.LambdaHack.Common.Point
 import           Game.LambdaHack.Common.State
 import qualified Game.LambdaHack.Common.Tile as Tile
 import           Game.LambdaHack.Common.Time
 import           Game.LambdaHack.Common.Types
 import           Game.LambdaHack.Common.Vector
 import qualified Game.LambdaHack.Content.ItemKind as IK
-import           Game.LambdaHack.Content.ModeKind
+import           Game.LambdaHack.Content.FactionKind
 import qualified Game.LambdaHack.Content.TileKind as TK
-import           Game.LambdaHack.Common.Point
 import qualified Game.LambdaHack.Definition.Ability as Ability
 import           Game.LambdaHack.Definition.Defs
 
@@ -138,13 +134,10 @@ calculateTotal fid s =
 
 -- | Price an item, taking count into consideration.
 itemPrice :: Int -> IK.ItemKind -> Int
-itemPrice jcount itemKind = case lookup "valuable" $ IK.ifreq itemKind of
+itemPrice jcount itemKind = case lookup IK.VALUABLE $ IK.ifreq itemKind of
   Just k -> jcount * k
   Nothing -> 0
 
-mergeItemQuant :: ItemQuant -> ItemQuant -> ItemQuant
-mergeItemQuant (k2, it2) (k1, it1) = (k1 + k2, it1 ++ it2)
-
 findIid :: ActorId -> FactionId -> ItemId -> State
         -> [(ActorId, (Actor, CStore))]
 findIid leader fid iid s =
@@ -153,54 +146,23 @@ findIid leader fid iid s =
         let itemsOfCStore store =
               let bag = getBodyStoreBag b store s
               in map (\iid2 -> (iid2, (aid, (b, store)))) (EM.keys bag)
-            stores = [CInv, CEqp, COrgan] ++ [CSha | aid == leader]
+            stores = [CEqp, COrgan] ++ [CStash | aid == leader]
         in concatMap itemsOfCStore stores
       items = concatMap itemsOfActor actors
   in map snd $ filter ((== iid) . fst) items
 
-combinedGround :: FactionId -> State -> ItemBag
-combinedGround fid s =
-  let bs = inline fidActorNotProjGlobalAssocs fid s
-  in EM.unionsWith mergeItemQuant
-     $ map (\(_, b) -> getFloorBag (blid b) (bpos b) s) bs
-
--- Trunk not considered (if stolen).
-combinedOrgan :: FactionId -> State -> ItemBag
-combinedOrgan fid s =
-  let bs = inline fidActorNotProjGlobalAssocs fid s
-  in EM.unionsWith mergeItemQuant $ map (borgan . snd) bs
-
-combinedEqp :: FactionId -> State -> ItemBag
-combinedEqp fid s =
-  let bs = inline fidActorNotProjGlobalAssocs fid s
-  in EM.unionsWith mergeItemQuant $ map (beqp . snd) bs
-
-combinedInv :: FactionId -> State -> ItemBag
-combinedInv fid s =
-  let bs = inline fidActorNotProjGlobalAssocs fid s
-  in EM.unionsWith mergeItemQuant $ map (binv . snd) bs
-
 -- Trunk not considered (if stolen).
 combinedItems :: FactionId -> State -> ItemBag
 combinedItems fid s =
-  let shaBag = gsha $ sfactionD s EM.! fid
+  let stashBag = getFactionStashBag fid s
       bs = map snd $ inline fidActorNotProjGlobalAssocs fid s
-  in EM.unionsWith mergeItemQuant $ map binv bs ++ map beqp bs ++ [shaBag]
-
-combinedFromLore :: SLore -> FactionId -> State -> ItemBag
-combinedFromLore slore fid s = case slore of
-  SItem -> combinedItems fid s
-  SOrgan -> combinedOrgan fid s
-  STrunk -> combinedOrgan fid s
-  SCondition -> combinedOrgan fid s
-  SBlast -> EM.empty
-  SEmbed -> EM.empty
+  in EM.unionsWith mergeItemQuant $ map beqp bs ++ [stashBag]
 
 getActorBody :: ActorId -> State -> Actor
 {-# INLINE getActorBody #-}
 getActorBody aid s = sactorD s EM.! aid
 
--- For now, faction and tactic skill modifiers only change
+-- For now, faction and doctrine skill modifiers only change
 -- the stats that affect permitted actions (@SkMove..SkApply@),
 -- so the expensive @actorCurrentSkills@ operation doesn't need to be used
 -- when checking the other skills, e.g., for FOV calculations,
@@ -214,11 +176,12 @@ actorCurrentSkills :: Maybe ActorId -> ActorId -> State -> Ability.Skills
 actorCurrentSkills mleader aid s =
   let body = getActorBody aid s
       actorMaxSk = getActorMaxSkills aid s
-      player = gplayer . (EM.! bfid body) . sfactionD $ s
-      skillsFromTactic = Ability.tacticSkills $ ftactic player
+      fact = (EM.! bfid body) . sfactionD $ s
+      skillsFromDoctrine = Ability.doctrineSkills $ gdoctrine fact
       factionSkills
         | Just aid == mleader = Ability.zeroSkills
-        | otherwise = fskillsOther player `Ability.addSkills` skillsFromTactic
+        | otherwise = fskillsOther (gkind fact)
+                      `Ability.addSkills` skillsFromDoctrine
   in actorMaxSk `Ability.addSkills` factionSkills
 
 -- Check that the actor can move, also between levels and through doors.
@@ -239,13 +202,8 @@ getCarriedAssocsAndTrunk :: Actor -> State -> [(ItemId, Item)]
 getCarriedAssocsAndTrunk b s =
   -- The trunk is important for a case of spotting a caught projectile
   -- with a stolen projecting item. This actually does happen.
-  let trunk = EM.singleton (btrunk b) (1, [])
-  in bagAssocs s $ EM.unionsWith const [binv b, beqp b, borgan b, trunk]
-
-getCarriedIidCStore :: Actor -> [(ItemId, CStore)]
-getCarriedIidCStore b =
-  let bagCarried (cstore, bag) = map (,cstore) $ EM.keys bag
-  in concatMap bagCarried [(CInv, binv b), (CEqp, beqp b), (COrgan, borgan b)]
+  let trunk = EM.singleton (btrunk b) quantSingle
+  in bagAssocs s $ EM.unionsWith const [beqp b, borgan b, trunk]
 
 getContainerBag :: Container -> State -> ItemBag
 getContainerBag c s = case c of
@@ -269,11 +227,15 @@ getBodyStoreBag b cstore s =
     CGround -> getFloorBag (blid b) (bpos b) s
     COrgan -> borgan b
     CEqp -> beqp b
-    CInv -> binv b
-    CSha -> gsha $ sfactionD s EM.! bfid b
+    CStash -> getFactionStashBag (bfid b) s
+
+getFactionStashBag :: FactionId -> State -> ItemBag
+getFactionStashBag fid s = case gstash $ sfactionD s EM.! fid of
+  Just (lid, pos) -> getFloorBag lid pos s
+  Nothing -> EM.empty
 
 mapActorItems_ :: Monad m
-               => (CStore -> ItemId -> ItemQuant -> m a) -> Actor -> State
+               => (CStore -> ItemId -> ItemQuant -> m ()) -> Actor -> State
                -> m ()
 mapActorItems_ f b s = do
   let notProcessed = [CGround]
@@ -283,13 +245,8 @@ mapActorItems_ f b s = do
         mapM_ (uncurry $ f cstore) $ EM.assocs bag
   mapM_ g sts
 
-getActorAssocs :: ActorId -> CStore -> State -> [(ItemId, Item)]
+getActorAssocs :: ActorId -> CStore -> State -> [(ItemId, (Item, ItemQuant))]
 getActorAssocs aid cstore s =
-  let b = getActorBody aid s
-  in bagAssocs s $ getBodyStoreBag b cstore s
-
-getActorAssocsK :: ActorId -> CStore -> State -> [(ItemId, (Item, ItemQuant))]
-getActorAssocsK aid cstore s =
   let b = getActorBody aid s
   in bagAssocsK s $ getBodyStoreBag b cstore s
 
@@ -340,18 +297,6 @@ actorInAmbient b s =
   let lvl = (EM.! blid b) . sdungeon $ s
   in Tile.isLit (coTileSpeedup $ scops s) (lvl `at` bpos b)
 
-canDeAmbientList :: Actor -> State -> [Point]
-canDeAmbientList b s =
-  let COps{coTileSpeedup} = scops s
-      lvl = (EM.! blid b) . sdungeon $ s
-      posDeAmbient p =
-        let t = lvl `at` p
-        in Tile.isWalkable coTileSpeedup t  -- no time to waste altering
-           && not (Tile.isLit coTileSpeedup t)
-  in if Tile.isLit coTileSpeedup (lvl `at` bpos b)
-     then filter posDeAmbient (vicinityUnsafe $ bpos b)
-     else []
-
 -- Check whether an actor can displace another. We assume they are adjacent
 -- and they are foes.
 dispEnemy :: ActorId -> ActorId -> Ability.Skills -> State -> Bool
@@ -363,12 +308,15 @@ dispEnemy source target actorMaxSk s =
         in any friend adjAssocs
       sb = getActorBody source s
       tb = getActorBody target s
-      dozes = bwatch tb `elem` [WSleep, WWake]
+      tfact = sfactionD s EM.! bfid tb
   in bproj tb
      || not (actorDying tb
              || actorWaits tb
              || Ability.getSk Ability.SkMove actorMaxSk <= 0
-                && not dozes  -- roots weak if the tree sleeps
+                  -- sometimes this comes from sleep, but it's transient
+                  -- and if we made exception for sleep, we would displace
+                  -- immobile sleeping actors
+             || Just (blid tb, bpos tb) == gstash tfact
              || hasBackup sb && hasBackup tb)  -- solo actors are flexible
 
 itemToFull :: ItemId -> State -> ItemFull
@@ -377,14 +325,14 @@ itemToFull iid s =
 
 fullAssocs :: ActorId -> [CStore] -> State -> [(ItemId, ItemFull)]
 fullAssocs aid cstores s =
-  let allAssocs = concatMap (\cstore -> getActorAssocsK aid cstore s) cstores
+  let allAssocs = concatMap (\cstore -> getActorAssocs aid cstore s) cstores
       iToFull (iid, (item, _kit)) =
         (iid, itemToFull6 (scops s) (sdiscoKind s) (sdiscoAspect s) iid item)
   in map iToFull allAssocs
 
 kitAssocs :: ActorId -> [CStore] -> State -> [(ItemId, ItemFullKit)]
 kitAssocs aid cstores s =
-  let allAssocs = concatMap (\cstore -> getActorAssocsK aid cstore s) cstores
+  let allAssocs = concatMap (\cstore -> getActorAssocs aid cstore s) cstores
       iToFull (iid, (item, kit)) =
         (iid, ( itemToFull6 (scops s) (sdiscoKind s) (sdiscoAspect s) iid item
               , kit ))
@@ -419,8 +367,19 @@ getItemKindServer item s = okind (coitem $ scops s) $ getItemKindIdServer item s
 getIidKindServer :: ItemId -> State -> IK.ItemKind
 getIidKindServer iid s = getItemKindServer (getItemBody iid s) s
 
--- | Determine the dungeon level of the container. If the item is in a shared
--- stash, the level depends on which actor asks.
+tileAlterable :: LevelId -> Point -> State -> Bool
+tileAlterable lid pos s =
+  let COps{coTileSpeedup} = scops s
+      embeds = getEmbedBag lid pos s
+      lvl = sdungeon s EM.! lid
+      t = lvl `at` pos
+      triggerable = any (\iid -> not $ null $ IK.ieffects $ getIidKind iid s)
+                        (EM.keys embeds)
+  in Tile.isModifiable coTileSpeedup t || triggerable
+
+-- | Determine the dungeon level of the container. If the item is in
+-- the shared stash, the level depends on which actor asks, not where
+-- the stash is located physically.
 lidFromC :: Container -> State -> LevelId
 lidFromC (CFloor lid _) _ = lid
 lidFromC (CEmbed lid _) _ = lid
@@ -433,6 +392,22 @@ posFromC (CEmbed _ pos) _ = pos
 posFromC (CActor aid _) s = bpos $ getActorBody aid s
 posFromC c@CTrunk{} _ = error $ "" `showFailure` c
 
+vicinityFoeAdj :: ((ActorId, Actor) -> Bool) -> ActorId -> State -> Bool
+{-# INLINE vicinityFoeAdj #-}
+vicinityFoeAdj predicate aid s =
+  let body = getActorBody aid s
+      lvl = (EM.! blid body) . sdungeon $ s
+      fact = (EM.! bfid body) . sfactionD $ s
+      f !p = case posToBigLvl p lvl of
+        Nothing -> False
+        Just aid2 -> let b2 = getActorBody aid2 s
+                     in isFoe (bfid body) fact (bfid b2)
+                        && predicate (aid2, b2)
+      h !p = case posToProjsLvl p lvl of
+        [] -> False
+        aid2 : _ -> isFoe (bfid body) fact . bfid $ getActorBody aid2 s
+  in any (\p -> f p || h p) $ vicinityUnsafe $ bpos body
+
 -- | Require that any non-dying foe is adjacent. We include even
 -- projectiles that explode when stricken down, because they can be caught
 -- and then they don't explode, so it makes sense to focus on handling them.
@@ -441,19 +416,10 @@ posFromC c@CTrunk{} _ = error $ "" `showFailure` c
 -- if the actor would need to flee instead of meleeing, but fleeing
 -- with *many* projectiles adjacent is a possible waste of a move anyway).
 anyFoeAdj :: ActorId -> State -> Bool
-anyFoeAdj aid s =
-  let body = getActorBody aid s
-      lvl = (EM.! blid body) . sdungeon $ s
-      fact = (EM.! bfid body) . sfactionD $ s
-      f !p = case posToBigLvl p lvl of
-        Nothing -> False
-        Just aid2 -> g $ getActorBody aid2 s
-      g !b = inline isFoe (bfid body) fact (bfid b)
-             && bhp b > 0  -- uncommon
-      h !p = case posToProjsLvl p lvl of
-        [] -> False
-        aid2 : _ -> g $ getActorBody aid2 s
-  in any (\ p -> f p || h p) $ vicinityUnsafe $ bpos body
+anyFoeAdj = vicinityFoeAdj (const True)
+
+anyHarmfulFoeAdj :: ActorMaxSkills -> ActorId -> State -> Bool
+anyHarmfulFoeAdj = vicinityFoeAdj . uncurry . actorWorthKilling
 
 adjacentBigAssocs :: Actor -> State -> [(ActorId, Actor)]
 {-# INLINE adjacentBigAssocs #-}
@@ -478,22 +444,31 @@ armorHurtBonus source target s =
       tMaxSk = getActorMaxSkills target s
   in armorHurtCalculation (bproj sb) sMaxSk tMaxSk
 
--- | Check if any non-dying foe (projectile or not) is adjacent
--- to any of our normal actors (whether they can melee or just need to flee,
--- in which case alert is needed so that they are not slowed down by others).
+-- | Check if any non-dying foe is adjacent to any of our normal actors
+-- and either can harm them via melee or can attack from a distance.
+-- Otherwise no point meleeing him. Projectiles are ignored, because
+-- they are not actively attempted to melee, see @meleeAny@.
+-- This is regardless of whether our actor can melee or just needs to flee,
+-- in which case alert is needed so that he is not slowed down by others.
+-- However, if our actor can't move nor melee, no real combat is taking place.
 -- This is needed only by AI and computed as lazily as possible.
-inMelee :: FactionId -> LevelId -> State -> Bool
-inMelee !fid !lid s =
+inMelee :: ActorMaxSkills -> FactionId -> LevelId -> State -> Bool
+inMelee !actorMaxSkills !fid !lid s =
   let fact = sfactionD s EM.! fid
-      f !b = blid b == lid
-             && inline isFoe fid fact (bfid b)  -- costly
-             && bhp b > 0  -- uncommon
-      allFoes = filter f $ EM.elems $ sactorD s
-      g !b = bfid b == fid
-             && blid b == lid
-             && not (bproj b)
-             && bhp b > 0
-      allOurs = filter g $ EM.elems $ sactorD s
+      f (!aid, !b) =
+        blid b == lid
+        && not (bproj b)
+        && inline isFoe fid fact (bfid b)  -- costly
+        && actorWorthKilling actorMaxSkills aid b
+      allFoes = filter f $ EM.assocs $ sactorD s
+      g (!aid, !b) = bfid b == fid
+                     && blid b == lid
+                     && not (bproj b)
+                     && bhp b > 0
+                     && let actorMaxSk = actorMaxSkills EM.! aid
+                        in Ability.getSk Ability.SkMove actorMaxSk > 0
+                           || actorCanMeleeToHarm actorMaxSkills aid b
+      allOurs = filter g $ EM.assocs $ sactorD s
       -- We assume foes are less numerous, even though they may come
       -- from multiple factions and they contain projectiles,
       -- because we see all our actors, while many foes may be hidden.
@@ -505,6 +480,7 @@ inMelee !fid !lid s =
       -- thus increasing allocation a bit, but not by much, because
       -- the set should be rather saturated.
       -- If there are no foes in sight, we don't iterate at all.
-      setFoeVicinity = ES.fromList $ concatMap (vicinityUnsafe . bpos) allFoes
+      setFoeVicinity =
+        ES.fromList $ concatMap (vicinityUnsafe . bpos . snd) allFoes
   in not (ES.null setFoeVicinity)  -- shortcut
-     && any (\b -> bpos b `ES.member` setFoeVicinity) allOurs
+     && any (\(_, b) -> bpos b `ES.member` setFoeVicinity) allOurs
